-----------------------------------------------------------------------------------------------
-- Client Lua Script for AAIPicker 
-- Copyright (c) NCsoft. All rights reserved
-----------------------------------------------------------------------------------------------
 
require "Window"
require "CraftingLib"
 
-----------------------------------------------------------------------------------------------
-- AAIPicker Module Definition
-----------------------------------------------------------------------------------------------
local AAIPicker = {} 
 
-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function AAIPicker:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 

    return o
end

function AAIPicker:Init()
	local bHasConfigureFunction = false
	local strConfigureButtonText = ""
	local tDependencies = {}
    Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end

-- Shows the window when launching the craft window.
function AAIPicker:showWindow(self)
	if not self.wndMain:IsShown() then
		self.wndMain:Show(true, true)
	end
end
 

-----------------------------------------------------------------------------------------------
-- AAIPicker OnLoad
-----------------------------------------------------------------------------------------------
function AAIPicker:OnLoad()
    -- load the form file
	self.xmlDoc = XmlDoc.CreateFromFile("AAIPicker.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
end

-----------------------------------------------------------------------------------------------
-- AAIPicker OnDocLoaded
-----------------------------------------------------------------------------------------------
function AAIPicker:OnDocLoaded()
	-- Successful load of the XML file?
	if self.xmlDoc ~= nil and self.xmlDoc:IsLoaded() then
	    self.wndMain = Apollo.LoadForm(self.xmlDoc, "AAIPickerForm", nil, self)
		-- Perhaps the XML form is malformed?
		if self.wndMain == nil then
			Apollo.AddAddonErrorText(self, "Could not load the main window for some reason.")
			return
		end
		
	    self.wndMain:Show(false)
		-- Register the GenericEvent_StartCraftingGrid,
		-- which is triggered when bringing up the crafting grid
		-- or when the user clicks on "Start Craft"
		Apollo.RegisterEventHandler("GenericEvent_StartCraftingGrid", "OnGenericEvent_StartCraftingGrid", self)
		
		-- Also register InitializeSchematicsTree to close the AAIP window.
		-- This event is triggered when the user brings up the schematics' list at the crafting table.
		Apollo.RegisterEventHandler("GenericEvent_InitializeSchematicsTree", "CloseAAIPWindow", self)
		
		-- This event will allow us to "save" the window's position, instead of putting it at the
		-- same place at each /reloadui...
		Apollo.RegisterEventHandler("WindowManagementReady", "OnWindowManagementReady", self)
		
		-- Handles the stuff related to "I clicked on 'Craft', but I want to automatically begin another one!"
		Apollo.RegisterTimerHandler("AAIP_CraftTimer", "OnAAIPCraftTimer", self)
		Apollo.CreateTimer("AAIP_CraftTimer", 3.25, false)
		Apollo.StopTimer("AAIP_CraftTimer")
	end
end

-- Triggered by the WindowManagementReady event.
-- Will allow us to load/save the position of the window between sessions.
function AAIPicker:OnWindowManagementReady()
	Event_FireGenericEvent("WindowManagementAdd", {wnd = self.wndMain, strName = "Additive AI Picker"})
end

-----------------------------------------------------------------------------------------------
-- AAIPicker Functions
-----------------------------------------------------------------------------------------------
-- Define general functions here

-- The sub-recipe's target X, Y, and radius
local targetItemX = 0
local targetItemY = 0
local targetItemR = 0

-- Stocks the best choice of the choice tree.
local bestChoice = nil
-- Stocks the best choice rate of success of the choice tree.
local bestChoiceRate = 0
-- How many additives we currently have?
local currentAdditiveCount = 0
-- The refresh timer.
local refreshTimer = nil
-- What's the current schematic's ID?
local currentIdSchematic = nil
-- What's the name of the recipe the user wants to craft?
local recipeName = ""
-- How many crafts we still have to do?
local remainingCrafts = 0

-- This function is triggered by GenericEvent_StartCraftingGrid, 
-- and called by OnWait if the additive number has changed during the craft
-- Parameters :
-- idSchematic : The identifier of the schematic, sent by the event...
function AAIPicker:OnGenericEvent_StartCraftingGrid(idSchematic)
	currentIdSchematic = idSchematic
	AAIPicker:showWindow(self) -- Show the window if it's not the case.
	
	-- Let's update the combo box...
	local comboBox = self.wndMain:FindChild("RecipeName")
	comboBox:DeleteAll()
	comboBox:AddItem("Manual mode")
	comboBox:AddItem("Default recipe")
	local tSchematicInfo = CraftingLib.GetSchematicInfo(currentIdSchematic)
	for idx, tCurrSubRecipe in pairs(tSchematicInfo.tSubRecipes) do
		comboBox:AddItem(tCurrSubRecipe.strName);
	end
	if comboBox:GetText() == "" then
		comboBox:SetText("Manual mode")
		comboBox:SelectItemByText("Manual mode")
		recipeName = "Manual mode"
	end
	
	-- Special case : there's no point trying to help the meatbag if he doesn't want to!
	if recipeName == "Manual mode" then
		Print("Manual mode currently selected.")
	-- Special case : If the input is 'Default recipe', it means the meat bag only wants to craft the default recipe.
	elseif recipeName == "Default recipe" then
		Print("Crafting the default recipe for this schematic.")
		local currentCraft = CraftingLib.GetCurrentCraft()
		-- Are we currently crafting something, or did we only bring up the crafting grid in preview mode?
		if currentCraft ~= nil then
			-- Let's complete the craft, as there's no additive to add!
			CraftingLib.CompleteCraft()
			Apollo.StartTimer("AAIP_CraftTimer")
		else
			AAIPicker:checkRemainingCraftingAttempts(self)
		end
	else
		AAIPicker:craftSubRecipe(self)
	end
end

-- This function is the most important part of AAIP, as it will handle the automation of a sub-recipe craft.
-- Parameters :
-- self : A link to self, allowing us to get the add-on's window.
function AAIPicker:craftSubRecipe(self)
	local tSchematicInfo = CraftingLib.GetSchematicInfo(currentIdSchematic)
	local subRecipe = nil
	
	-- Let's ensure we find the sub-recipe.
		
	-- First, let's try to find a sub-recipe with the EXACT name.
	for idx, tCurrSubRecipe in pairs(tSchematicInfo.tSubRecipes) do
		local item = tCurrSubRecipe.itemOutput	
		if string.lower(tCurrSubRecipe.strName) == string.lower(recipeName) then 
			recipeName = tCurrSubRecipe.strName
			targetItemX = tCurrSubRecipe.fVectorX 
			targetItemY = tCurrSubRecipe.fVectorY
			targetItemR = tCurrSubRecipe.fRadius
			subRecipe = tCurrSubRecipe
			-- We can't do crap if the sub-recipe is unknown or locked...
			if subRecipe.bIsUndiscovered or not subRecipe.bIsKnown then
				Print("Found sub-recipe with name '" .. recipeName .. "', but it is unknown or locked. Discover it first!")
				subRecipe = nil
				return
			end
			break
		end
	end
		
	-- Did we found the sub-recipe? If there's no sub-recipe with the provided name, let the meatbag do its own stuff.
	if subRecipe ~= nil then
	
		if self.wndMain:FindChild("ShowDebugButton"):IsChecked() then
			Print("Found sub-recipe with name '" .. recipeName .. "'")
		end
		
		local currentCraft = CraftingLib.GetCurrentCraft()
		-- Are we currently crafting something, or did we only bring up the crafting grid in preview mode?
		if currentCraft ~= nil then
			local additiveList = CraftingLib.GetAvailableAdditives(tSchematicInfo.eTradeskillId, currentIdSchematic)
			currentAdditiveCount = currentCraft.nAdditiveCount
			
			-- Perhaps we don't even need to do anything...
			local craftAlreadySuccessful = AAIPicker:calculateDistanceFromTwoPoints(targetItemX, targetItemY, currentCraft.fVectorX, currentCraft.fVectorY) < targetItemR
			if not craftAlreadySuccessful then
				-- Let's build the choice tree to find the most optimal route.
				local choiceTree = AAIPicker:buildChoiceTree(nil, currentAdditiveCount, tSchematicInfo.nMaxAdditives, additiveList, currentCraft.fVectorX, currentCraft.fVectorY, 0)
				-- Read it to get the most interesting choice
				AAIPicker:readChoiceTree(choiceTree)
				-- Do we have some chance to have a success?
				if bestChoice ~= nil then
					-- Inform the user...
					if self.wndMain:FindChild("ShowDebugButton"):IsChecked() then
						Print("Best choice has " .. math.floor(bestChoiceRate) .. "% chances to succeed")
						Print(AAIPicker:printAdditives(bestChoice))
					end
					-- Put da additive in da stew!
					CraftingLib.AddAdditive(AAIPicker:getFirstAdditive(bestChoice))
					
					-- Reset the best choice parameters for the next call.
					bestChoice = nil
					bestChoiceRate = 0
				
					-- Trigger an event, which will check when we'll be able to recalculate the most optimal path
					refreshTimer = ApolloTimer.Create(0.05, false, "OnWait", self)
					refreshTimer:Start()
				elseif currentAdditiveCount == tSchematicInfo.nMaxAdditives then
					if self.wndMain:FindChild("ShowDebugButton"):IsChecked() then
						Print("Can't add more additives.")
					end
					-- Reset the best choice parameters for the next call.
					bestChoice = nil
					bestChoiceRate = 0
					-- Let's complete the craft, we're screwed for this one.
					CraftingLib.CompleteCraft()
					Apollo.StartTimer("AAIP_CraftTimer")
				else
					if self.wndMain:FindChild("ShowDebugButton"):IsChecked() then
						Print("There's not a single chance to succeed for this sub-recipe...")
					end
					
					-- Reset the best choice parameters for the next call.
					bestChoice = nil
					bestChoiceRate = 0
					-- Bah...Let's complete the craft.
					CraftingLib.CompleteCraft()
					Apollo.StartTimer("AAIP_CraftTimer")
				end
			else
				if self.wndMain:FindChild("ShowDebugButton"):IsChecked() then
					Print("Craft is already successful, no need to add more additives!")
				end
				-- Bingo! Let's complete the craft!
				CraftingLib.CompleteCraft()
				Apollo.StartTimer("AAIP_CraftTimer")
			end
		else
			AAIPicker:checkRemainingCraftingAttempts(self)
		end
	else
		Print("Did not found an unlocked sub-recipe with name '" .. recipeName .. "'")
	end
end

-- This function checks the amount of crafting attempts.
-- If we still have stuff to craft, Start a craft, and trigger a StartCraftingGrid event.
-- If not...Well, do nothing!
-- Parameters :
-- self : A link to self, to access the add-on's window.
function AAIPicker:checkRemainingCraftingAttempts(self)
	remainingCrafts = remainingCrafts - 1
	
	if remainingCrafts >= 0 then
		self.wndMain:FindChild("MultiCraftNumber"):SetText(remainingCrafts)
		-- "Click" on "Start Craft".
		CraftingLib.CraftItem(currentIdSchematic, nil) 
		-- Trigger the StartCraftingGrid event to really begin the craft.
		Event_FireGenericEvent("GenericEvent_StartCraftingGrid", currentIdSchematic) 
	else
		Print("No more crafts!")
	end
end

-- A function to wait for the availability of the crafting grid.
-- When the additive count returned by CraftingLib is different from the one we currently have in stock,
-- the function will stop calling itself, and trigger another round of best-path calculation.
function AAIPicker:OnWait()
	local currentCraft = CraftingLib.GetCurrentCraft()
	if currentCraft ~= nil then
		local additiveCount = currentCraft.nAdditiveCount
		-- Do we continue to do another call to OnWait?
		if additiveCount == currentAdditiveCount then 
			self.refreshTimer = ApolloTimer.Create(0.05, false, "OnWait", self)
			self.refreshTimer:Start() 
			return
		end
		-- If we did not enter the 'if' statement, it means we can go on.
		Event_FireGenericEvent("GenericEvent_StartCraftingGrid", currentIdSchematic)
	end
end

-- This function is triggered by AAIP_CraftTimer, normally ~1s after completing a craft.
function AAIPicker:OnAAIPCraftTimer()
	Apollo.StopTimer("AAIP_CraftTimer")
	Event_FireGenericEvent("GenericEvent_StartCraftingGrid", currentIdSchematic)
end

-- This function gets the first additive of a choice tree path.
-- Parameters :
-- treeItem - The tree item representing the path.
-- Returns :
-- The itemData representing the first additive of the path.
function AAIPicker:getFirstAdditive(treeItem)
	-- If the treeItem has no more parents, it means we're on the first level of the tree, so return the additive.
	if treeItem.parent == nil then
		return treeItem.additive
	else
		-- If not, it means we need to go to the parent element.
		return AAIPicker:getFirstAdditive(treeItem.parent)
	end
end

-- This function will print the additives' list of a choice tree path, with the price of each additive.
-- Parameters :
-- treeItem - The tree item representing the path.
-- Returns :
-- A string representing the additives path from the beginning of the path to the end of it and their prices
function AAIPicker:printAdditives(treeItem)
	if treeItem == nil then
		-- There's no more tree items, just return "".
		return ""
	else
		-- Another recursive function. As the treeItem we initially give is at lower levels of the tree
		-- We build the string from the left, adding parents' additives on the right side.
		return AAIPicker:printAdditives(treeItem.parent) .. " > " .. treeItem.additive:GetName() .. " (" .. AAIPicker:HelperStringMoneyConvert(treeItem.additive:GetBuyPrice():GetAmount()) .. ")"
	end
end

-- Reads the choice tree to determine the most optimal path.
-- It updates bestChoice and bestChoiceRate variables.
-- Parameters :
-- choiceTree - The choice tree we must read through.
function AAIPicker:readChoiceTree(choiceTree)
	for i = 1, table.getn(choiceTree) do
		-- Additive circle is entirely contained by the target item's circle.
		if choiceTree[i].remainingDistance + choiceTree[i].cumulatedRadius < targetItemR then 
			bestChoice = choiceTree[i]
			bestChoiceRate = 100
			return nil
		-- Both circles intersect on two points, or additives' circle contains the target circle : we can miss, calculate the success rate...
		elseif choiceTree[i].remainingDistance - choiceTree[i].cumulatedRadius < targetItemR or choiceTree[i].remainingDistance + targetItemR < choiceTree[i].cumulatedRadius then
			-- Let's see how much space we have in common...
			local chanceToSucceed = AAIPicker:calculateSuccessChance(choiceTree[i].x, choiceTree[i].y, choiceTree[i].cumulatedRadius)
			-- Is it a better choice than the one we already stocked?
			if bestChoiceRate < chanceToSucceed then
				bestChoice = choiceTree[i]
				bestChoiceRate = chanceToSucceed 
			end
			-- Continue by reading the subTree...
			AAIPicker:readChoiceTree(choiceTree[i].subTree)
		-- Literally 0% chance : Both circles are separated.
		else
			-- Continue by reading the subTree...
			AAIPicker:readChoiceTree(choiceTree[i].subTree)
		end
		
	end
end

-- This function aims to use good old maths to calculate the success chance of a particular additive.
-- Only used in two cases : Both circles intersect on two points, or additives' circle contains the target circle.
-- Parameters :
-- additiveCircleX - The abcissa of the additives' circle.
-- additiveCircleY - The ordinate of the additives' circle.
-- additiveCircleR - The radius of the additives' circle.
-- Returns :
-- A success rate between 0.0 and 99.9999...
function AAIPicker:calculateSuccessChance(additiveCircleX, additiveCircleY, additiveCircleR)
	
	-- Determine the straight-line distance between the centers
	local d = AAIPicker:calculateDistanceFromTwoPoints(targetItemX, targetItemY, additiveCircleX, additiveCircleY)
	local r = targetItemR
	local R = additiveCircleR
	
	-- If the additives' circle contains the target...
	if d + r < R then
		return (r * r) / (R * R) * 100.0 -- This one is simple : pi*r / pi*R ... Pi is simplified on both sides.
	else 
		-- Circles intersect. We must calculate the area of this intersection, and compare it to the additives' circle area.
		-- I've searched for formulas, so here's the thread I've stumbled upon...
		-- http://mathforum.org/library/drmath/view/54785.html
		local CBD = 2 * math.acos((R*R + d*d - r*r)/(2*R*d))
		local CAD = 2 * math.acos((r*r + d*d - R*R)/(2*r*d))
		
		local area = 0.5 * (CBD * R*R 
						  - R*R * math.sin(CBD) 
						  + CAD * r*r 
						  - r*r * math.sin(CAD))
		return area / (math.pi * additiveCircleR * additiveCircleR) * 100.0
	end
end

-- This function builds the choice tree, which will be used later to determine the best path to take.
-- Parameters :
-- parentTree - The parent element of the choiceTree we're building. Can be nil (1st level elements have no parents...).
-- currentDepth - How deep we are in the tree.
-- maxDepth - How deep we want to stop building the tree (the max additive count)
-- currentX - The starting point's abcissa. We need it because we build the best path from this point.
-- currentY - The starting point's ordinate. We need it because we build the best path from this point.
-- cumulatedR - As we're using additives to estimate the best path, we must take their radiuses into account.
-- Returns :
-- A complete choice tree, representing the current problem.
function AAIPicker:buildChoiceTree(parentTree, currentDepth, maxDepth, additiveList, currentX, currentY, cumulatedR)
	-- We won't go deeper!
	if currentDepth >= maxDepth then
		return {}
	end
	
	local toReturn = {}
	local i = 1
	
	-- Let's get the distance from the recipe first.
	local distanceFromRecipe = AAIPicker:calculateDistanceFromTwoPoints(targetItemX, targetItemY, currentX, currentY)
	
	-- For each possible additive...
	for id, item in pairs(additiveList) do
		local tAdditive = item:GetAdditiveInfo()
		local craftAfterAdditiveX = currentX + tAdditive.fVectorX
		local craftAfterAdditiveY = currentY + tAdditive.fVectorY
		local cumulatedRadiusAfterAdditive = cumulatedR + tAdditive.fRadius 
		-- Get the distance from the recipe if we use this additive.
		local distanceFromRecipeWithAdditive = AAIPicker:calculateDistanceFromTwoPoints(targetItemX, targetItemY, craftAfterAdditiveX, craftAfterAdditiveY)
		-- Would it improve our route, even slightly?
		if distanceFromRecipeWithAdditive < distanceFromRecipe or currentDepth == 0 then
			-- Build the tree element. It has a parent, the item representing the additive,
			-- a position, the remaining distance from the target's center, the cumulated radius of additives
			-- and a subtree.
			toReturn[i] = {	parent = parentTree,
							additive = item, 
							x = craftAfterAdditiveX, 
							y = craftAfterAdditiveY, 
							remainingDistance = distanceFromRecipeWithAdditive, 
							cumulatedRadius = cumulatedRadiusAfterAdditive,
							subTree = {}}
			-- Build the subTree of this element by recursively calling buildChoiceTree...
			toReturn[i].subTree = AAIPicker:buildChoiceTree(toReturn[i], currentDepth + 1, maxDepth, additiveList, craftAfterAdditiveX, craftAfterAdditiveY, cumulatedRadiusAfterAdditive)
			i = i + 1
		end
	end
	
	return toReturn
end

-- Calculates the distance between two points A and B in a cartesian plane.
-- For the record, it is square root((XA-XB) + (YA-YB))...
-- Parameters :
-- XA : The abcissa of point A
-- YA : The ordinate of point A
-- XB : The abcissa of point B
-- YB : The ordinate of point B
-- Returns :
-- The distance between these two points as a float.
function AAIPicker:calculateDistanceFromTwoPoints(XA, YA, XB, YB)
	-- (XA-XB)...
	local distFromTargetX = math.pow(XA - XB, 2)
	-- (YA-YB)...
	local distFromTargetY = math.pow(YA - YB, 2)
	-- sqrt of these...
	return math.sqrt(distFromTargetX + distFromTargetY)
end

-- Builds a string to give a meatbag-friendly representation of a price.
-- Parameters :
-- nInCopper : An integer representing the price of an item in coppers.
-- Returns :
-- The price in a meatbag-friendly way, ie 12345678 will convert to 12p34g56s78c.
function AAIPicker:HelperStringMoneyConvert(nInCopper)
	local strResult = ""
	if nInCopper >= 1000000 then
		strResult = math.floor(nInCopper/1000000) .. "p"
	end
	if nInCopper >= 10000 then
		strResult = strResult .. math.floor(nInCopper % 1000000 / 10000) .. "g"
	end
	if nInCopper >= 100 then
		strResult = strResult .. math.floor(nInCopper % 10000 / 100) .. "s"
	end
	strResult = strResult .. math.floor(nInCopper % 100) .. "c"
	return strResult
end

-----------------------------------------------------------------------------------------------
-- AAIPickerForm Functions
-----------------------------------------------------------------------------------------------
-- when the OK button is clicked
function AAIPicker:OnOK(wndHandler, wndControl, eMouseButton)
	-- Get the recipe name entered by the user...
	recipeName = self.wndMain:FindChild("RecipeName"):GetText()
	-- Ensure that a number is given to the number of crafting attempts...
	if tonumber(self.wndMain:FindChild("MultiCraftNumber"):GetText()) ~= nil then
		remainingCrafts = tonumber(self.wndMain:FindChild("MultiCraftNumber"):GetText())
	else
		Print("ERROR : You must enter a number in the 'Crafting attempts' box, cupcake.")
		self.wndMain:FindChild("MultiCraftNumber"):SetText("0")
	end
	-- Try to help the meatbag to complete his craft
	if currentIdSchematic ~= nil then
		Event_FireGenericEvent("GenericEvent_StartCraftingGrid", currentIdSchematic)
	end
end

-- When the "Clear" button is clicked
function AAIPicker:OnClear(wndHandler, wndControl, eMouseButton)
	self.wndMain:FindChild("RecipeName"):SetText("")
	recipeName = ""
	self.wndMain:FindChild("MultiCraftNumber"):SetText("1")
	remainingCrafts = 0
	if currentIdSchematic ~= nil then
		Event_FireGenericEvent("GenericEvent_StartCraftingGrid", currentIdSchematic)
	end
end

-- When the user brings up the schematics' list...
function AAIPicker:CloseAAIPWindow(wndHandler, wndControl)
	self.wndMain:FindChild("RecipeName"):SetText("")
	recipeName = ""
	self.wndMain:FindChild("MultiCraftNumber"):SetText("1")
	remainingCrafts = 0
	self.wndMain:Close() -- hide the window
end

-- This function is triggered by clicking on "Max Attempts"
-- It will calculate how many attempts we can do for a given schematic, and put it in the
-- "Crafting attempts" textbox.
function AAIPicker:CalculateMaxAttempts(wndHandler, wndControl, eMouseButton)
	if currentIdSchematic ~= nil then
		local tSchematicInfo = CraftingLib.GetSchematicInfo(currentIdSchematic)
		local maxAttempts = 9001 -- Over 9000, lawl
		for idx, tMaterial in pairs(tSchematicInfo.tMaterials) do
			if  math.floor(tMaterial.itemMaterial:GetBackpackCount() / tMaterial.nAmount) < maxAttempts then
				maxAttempts = math.floor(tMaterial.itemMaterial:GetBackpackCount() / tMaterial.nAmount)
			end
		end
		self.wndMain:FindChild("MultiCraftNumber"):SetText(maxAttempts)
		remainingCrafts = maxAttempts
	else
		Print("ERROR : Open a schematic first, cupcake!")
	end
end

-----------------------------------------------------------------------------------------------
-- AAIPicker Instance
-----------------------------------------------------------------------------------------------
local AAIPickerInst = AAIPicker:new()
AAIPickerInst:Init()
